// This class implements a forward iterator using the circular list of pointers to define a
// subset of the population. Example of use:
//       CircConstIterAsForwardConstIter<&msl::Individual::family_next>
template<msl::Individual* msl::Individual::*next_ptr>
struct CircConstIterAsForwardConstIter
{
	msl::RefConst operator*() const { return *ptr_; }
	// As the iterator is incremented, the start state becomes false. Then iterators can compare
	// equal.
	CircConstIterAsForwardConstIter operator++()
	{
		ptr_ = ptr_->*next_ptr;
		if (start_)
			start_ = false;
		return *this;
	}
	// If neither iterator has been incremented, the iterators are differents.
	bool operator!=(CircConstIterAsForwardConstIter const& other) const
	{
		return (start_ && other.start_) || (ptr_ != other.ptr_);
	}
	// Used in ''min_element'' algorithm.
	bool operator==(CircConstIterAsForwardConstIter const& other) const { return !(*this != other); }
	msl::Individual const* ptr_;
	bool start_;
};

template<msl::Individual* msl::Individual::*next_ptr>
struct CircConstIterAsForwardConstIterWrapper
{
	CircConstIterAsForwardConstIter<next_ptr> begin() const { return { begin_, true }; }
	CircConstIterAsForwardConstIter<next_ptr> end() const { return { begin_, true }; }
	CircConstIterAsForwardConstIter<next_ptr> cbegin() const { return { begin_, true }; }
	CircConstIterAsForwardConstIter<next_ptr> cend() const { return { begin_, true }; }

	// Example of use : head.over_family().sum([](msl::RefConst ref) { return ref.x + ref.y; })
	double sum(msl::Lambda const& get) const
	{
		return std::accumulate(
		  this->cbegin(), this->cend(), double{}, [&](double sum, msl::RefConst ref) {
			  double const x{ get(ref) };
			  // Check if the value is not a missing one. Display an error and exit.
			  if (std::isnan(x))
				  msl::error_get_nan_interest_variable("sum", "<lambda>", ref);
			  return sum + x;
		  });
	}
	// Defined in the file ''include-circular-iterator-2.cpp''.
	double sum(msl::Table const& get) const;
	// Example of use : head.over_family().min([](msl::RefConst ref) { return ref.x + ref.y; })
	double min(msl::Lambda const& get) const
	{
		return get(
		  *std::min_element(this->cbegin(), this->cend(), [&](msl::RefConst ref1, msl::RefConst ref2) {
			  double const x1{ get(ref1) };
			  // Check if the value is not a missing one. Display an error and exit.
			  if (std::isnan(x1))
				  msl::error_get_nan_interest_variable("min", "<lambda>", ref1);
			  double const x2{ get(ref2) };
			  // Check if the value is not a missing one. Display an error and exit.
			  if (std::isnan(x2))
				  msl::error_get_nan_interest_variable("min", "<lambda>", ref2);
			  return x1 < x2;
		  }));
	}
	// Defined in the file ''include-circular-iterator-2.cpp''.
	double min(msl::Table const& get) const;
	// Example of use : head.over_family().ref_min([](msl::RefConst ref) { return ref.x + ref.y; })
	msl::RefConst ref_min(msl::Lambda const& get) const
	{
		return *std::min_element(
		  this->begin(), this->end(), [&](msl::RefConst ref1, msl::RefConst ref2) {
			  double const x1{ get(ref1) };
			  // Check if the value is not a missing one. Display an error and exit.
			  if (std::isnan(x1))
				  msl::error_get_nan_interest_variable("ref_min", "<lambda>", ref1);
			  double const x2{ get(ref2) };
			  // Check if the value is not a missing one. Display an error and exit.
			  if (std::isnan(x2))
				  msl::error_get_nan_interest_variable("ref_min", "<lambda>", ref2);
			  return x1 < x2;
		  });
	}
	// Defined in the file ''include-circular-iterator-2.cpp''.
	msl::RefConst ref_min(msl::Table const& get) const;
	// Example of use : head.over_family().max([](msl::RefConst ref) { return ref.x + ref.y; })
	double max(msl::Lambda const& get) const
	{
		return get(
		  *std::max_element(this->cbegin(), this->cend(), [&](msl::RefConst ref1, msl::RefConst ref2) {
			  double const x1{ get(ref1) };
			  // Check if the value is not a missing one. Display an error and exit.
			  if (std::isnan(x1))
				  msl::error_get_nan_interest_variable("max", "<lambda>", ref1);
			  double const x2{ get(ref2) };
			  // Check if the value is not a missing one. Display an error and exit.
			  if (std::isnan(x2))
				  msl::error_get_nan_interest_variable("max", "<lambda>", ref2);
			  return x1 < x2;
		  }));
	}
	// Defined in the file ''include-circular-iterator-2.cpp''.
	double max(msl::Table const& get) const;
	// Example of use : head.over_family().ref_max([](msl::RefConst ref) { return ref.x + ref.y; })
	msl::RefConst ref_max(msl::Lambda const& get) const
	{
		return *std::max_element(
		  this->begin(), this->end(), [&](msl::RefConst ref1, msl::RefConst ref2) {
			  double const x1{ get(ref1) };
			  // Check if the value is not a missing one. Display an error and exit.
			  if (std::isnan(x1))
				  msl::error_get_nan_interest_variable("ref_max", "<lambda>", ref1);
			  double const x2{ get(ref2) };
			  // Check if the value is not a missing one. Display an error and exit.
			  if (std::isnan(x2))
				  msl::error_get_nan_interest_variable("ref_max", "<lambda>", ref2);
			  return x1 < x2;
		  });
	}
	// Defined in the file ''include-circular-iterator-2.cpp''.
	msl::RefConst ref_max(msl::Table const& get) const;
	// Example of use : head.over_family().avg([](msl::RefConst ref) { return ref.x + ref.y; })
	double avg(msl::Lambda const& get) const
	{
		auto const pair{ std::accumulate(this->cbegin(),
			                               this->cend(),
			                               std::pair<double, size_t>{},
			                               [&](std::pair<double, size_t> pair, msl::RefConst ref) {
			                                 double const x{ get(ref) };
			                                 // Check if the value is not a missing one. Display an error
			                                 // and exit.
			                                 if (std::isnan(x))
				                                 msl::error_get_nan_interest_variable(
				                                   "avg", "<lambda>", ref);
			                                 return std::make_pair(pair.first + x, ++pair.second);
			                               }) };
		return pair.second == 0 ? std::numeric_limits<double>::quiet_NaN() : pair.first / pair.second;
	}
	// Defined in the file ''include-circular-iterator-2.cpp''.
	double avg(msl::Table const& get) const;
	msl::Individual const* begin_;
};

// The non const version.

template<msl::Individual* msl::Individual::*next_ptr>
struct CircIterAsForwardIter
{
	msl::Ref operator*() const { return *ptr_; }
	CircIterAsForwardIter operator++()
	{
		ptr_ = ptr_->*next_ptr;
		if (start_)
			start_ = false;
		return *this;
	}
	bool operator!=(CircIterAsForwardIter const& other) const
	{
		return (start_ && other.start_) || (ptr_ != other.ptr_);
	}

	msl::Individual* ptr_;
	bool start_;
};

template<msl::Individual* msl::Individual::*next_ptr>
struct CircIterAsForwardIterWrapper
{
	CircIterAsForwardIter<next_ptr> begin() { return { begin_, true }; }
	CircIterAsForwardIter<next_ptr> end() { return { begin_, true }; }
	CircIterAsForwardIter<next_ptr> cbegin() const { return { begin_, true }; }
	CircIterAsForwardIter<next_ptr> cend() const { return { begin_, true }; }

	double sum(msl::Lambda const& get) const
	{
		// The wrapper is converted to a const wrapper and the member function of the const iterator is
		// used.
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		return const_iterator_wrapper.sum(get);
	}
	double sum(msl::Table const& get) const
	{
		// The wrapper is converted to a const wrapper and the member function of the const iterator is
		// used.
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		return const_iterator_wrapper.sum(get);
	}
	double min(msl::Lambda const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		return const_iterator_wrapper.min(get);
	}
	double min(msl::Table const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		return const_iterator_wrapper.min(get);
	}
	msl::Ref ref_min(msl::Lambda const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		msl::RefConst ref_const{ const_iterator_wrapper.ref_min(get) };
		return const_cast<msl::Ref>(ref_const);
	}
	msl::Ref ref_min(msl::Table const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		msl::RefConst ref_const{ const_iterator_wrapper.ref_min(get) };
		return const_cast<msl::Ref>(ref_const);
	}
	double max(msl::Lambda const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		return const_iterator_wrapper.max(get);
	}
	double max(msl::Table const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		return const_iterator_wrapper.max(get);
	}
	msl::Ref ref_max(msl::Lambda const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		msl::RefConst ref_const{ const_iterator_wrapper.ref_max(get) };
		return const_cast<msl::Ref>(ref_const);
	}
	msl::Ref ref_max(msl::Table const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		msl::RefConst ref_const{ const_iterator_wrapper.ref_max(get) };
		return const_cast<msl::Ref>(ref_const);
	}
	double avg(msl::Lambda const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		return const_iterator_wrapper.avg(get);
	}
	double avg(msl::Table const& get) const
	{
		CircConstIterAsForwardConstIterWrapper<next_ptr> const_iterator_wrapper{ this->begin_ };
		return const_iterator_wrapper.avg(get);
	}
	msl::Individual* begin_;
};

// The generic function which returns the size of a subset from the head of this subset.
// ''head_ptr” is a pointer of the function member which returns the head flag.
// ''next_ptr” is the data member which gives the next element in the subset.
// Example of use:
//       generic_subset_size<&msl::Individual::family_head, &msl::Individual::family_next>(head)
template<bool (msl::Individual::*head_ptr)() const, msl::Individual* msl::Individual::*next_ptr>
inline size_t
generic_subset_size(msl::RefConst head)
{
	assert((head.*head_ptr)());
	size_t n{ 1 };
	auto ptr{ &head };
	while (ptr->*next_ptr != &head)
		ptr = ptr->*next_ptr, ++n;
	return n;
}

// The generic function which returns the head of a subset from a element of this subset.
// ''head_ptr” is a pointer of the function member which returns the head flag.
// ''next_ptr” is the data member which gives the next element in the subset.
// Example of use:
//       generic_get_subset_head<&msl::Individual::family_head, &msl::Individual::family_next>(head)
template<bool (msl::Individual::*head_ptr)() const, msl::Individual* msl::Individual::*next_ptr>
inline msl::Ref
generic_get_subset_head(msl::Ref ref, char const* f_name)
{
	auto ptr{ &ref };
	while (true) {
		if ((ptr->*head_ptr)())
			return *ptr;
		ptr = ptr->*next_ptr;
		if (ptr == &ref) {
			std::cerr << "No head found for " << f_name << '.' << std::endl;
			exit(1);
		}
	}
	return ref; /* NOT REACHED */
}

// The const version.
template<bool (msl::Individual::*head_ptr)() const, msl::Individual* msl::Individual::*next_ptr>
inline msl::RefConst
generic_get_subset_head(msl::RefConst ref, char const* f_name)
{
	auto ptr{ &ref };
	while (true) {
		if ((ptr->*head_ptr)())
			return *ptr;
		ptr = ptr->*next_ptr;
		if (ptr == &ref) {
			std::cerr << "No head found for " << f_name << '.' << std::endl;
			exit(1);
		}
	}
	return ref; /* NOT REACHED */
}

// The generic fonction which returns the forward iterator over a subset of the population.
// ''next_ptr” is the data member which gives the next element in the subset.
// Example of use:
//       generic_over_subset<&msl::Individual::family_next>(head)
template<msl::Individual* msl::Individual::*next_ptr>
auto
generic_over_subset(msl::Ref ref)
{
	return CircIterAsForwardIterWrapper<next_ptr>{ &ref };
}
// The const version.
template<msl::Individual* msl::Individual::*next_ptr>
auto
generic_over_subset(msl::RefConst ref)
{
	return CircConstIterAsForwardConstIterWrapper<next_ptr>{ &ref };
}
