#ifndef MSL_HPP
#define MSL_HPP

#include "fd-read-xlsx.hpp"
#include "fd-write-xlsx.hpp"
#include <algorithm>
#include <bitset>
#include <cassert>
#include <chrono>
#include <cmath>
#include <cstdlib>
#include <exception>
#include <execution>
#include <filesystem>
#include <fstream>
#include <functional>
#include <iostream>
#include <limits>
#include <map>
#include <memory>
#include <mutex>
#include <numeric>
#include <random>
#include <set>
#include <stdexcept>
#include <thread>
#include <vector>
#include <zlib.h>

#define MSL_SHOW(prm)                                                                              \
	std::cout << "msl library macro “MSL_SHOW” output - " #prm "{" << (prm) << '}' << std::endl;
#define MSL_LMBD(expression) [&](msl::RefConst ref) { return expression; }
#define MSL_FNCTN(expression) [&](msl::Ref ref) { expression; }

// Just by laziness...
namespace std {
using str = std::string;
}

namespace msl {

// Number of cores on the current computer.
static auto const THREADS_NR{ std::thread::hardware_concurrency() };
// A complicated way to get 255... The categorical variables are represented as uint8_t.
static auto constexpr CTG_MAX{ static_cast<size_t>(std::numeric_limits<uint8_t>::max()) };
// The maximum number of values for a CTG variable.
static auto constexpr CTG_NR{ CTG_MAX + 1 };

// This variable is defined in the “msl.cpp” file.
extern std::mutex global_lock;

// Forward declarations.
struct Individual;
namespace var {
namespace internals {
struct Table;
}
}

// Convenient types.
using Ref = msl::Individual&;
using RefConst = msl::Individual const&;
// To use with parallelised “Population::for_each”.
using Function = std::function<void(msl::Ref)>;
// Raw pointer of a free function returning a “float” from a const reference on a individual.
using Flt = float (*)(msl::RefConst);
// Raw pointer of a free function returning a “uint8_t” from a const reference on a individual.
using Ctg = uint8_t (*)(msl::RefConst);
// Lambda returning a “double” from a const reference on a individual.
using Lambda = std::function<double(msl::RefConst)>;
// Table defined for each variable of the model to give some informations on the variable.
using Table = msl::var::internals::Table;
// Array used in the sum by reduce to compute the frequences of each values of a “CTG” variable.
using SumByFreq = std::array<double, msl::CTG_NR>;
// Map to get the format associated with a “CTG” variable.
using FormatMap = std::pair<std::str, std::map<uint8_t, std::str>>;

// Declare the formats map as a “extern” value; the formats map is first initialised in the
// “msl.cpp” file; the function “load_formats” will set the actual values.
extern std::map<msl::Ctg, msl::FormatMap> global_formats;

// Load the parameters from a excel workbook, selecting a column with the “year” label.
void
load_parameters(std::str const& file_name, size_t year);

void
load_formats(std::str const& file_name);

struct Tabular : public std::map<size_t, std::pair<std::str, double>>
{
	// Save a ``Tabular'' in a excel workbook.
	void save(std::str const&) const;
	// Save two ``Tabular'' in a excel workbook.
	void save(msl::Tabular const&, std::str const&) const;
};

class Exception : public std::exception
{

public:
	Exception(std::str const& msg)
	  : msg_("msl library: " + msg + ((*crbegin(msg) == '?') ? ' ' : '.'))
	{
	}
	char const* what() const throw() { return msg_.c_str(); }

private:
	std::str const msg_;
};

// Declare an instance of std::vector<T> and reserve a capacity of n for it.
template<typename T>
std::vector<T>
reserve(size_t n)
{
	std::vector<T> rvo;
	rvo.reserve(n);
	return rvo;
}
// Call the member function “join” on each thread. Declare it as “inline” to bypass the “On
// Declaration Rule”.
inline void
join(std::vector<std::thread>& threads)
{
	for (auto& thread : threads)
		thread.join();
}

// Like the python enumerate function. To be used like that:
//    for (auto [i, ref] : msl::enumerate(iterable) ) { ... }
// “i” is the value of the index and “ref” is a reference on the current element.
// Source : https://www.reedbeta.com/blog/python-like-enumerate-in-cpp17/
template<typename T,
         typename Iterator = decltype(std::begin(std::declval<T>())),
         typename = decltype(std::end(std::declval<T>()))>
constexpr auto
enumerate(T&& iterable)
{
	struct DoubleIterator
	{
		bool operator!=(const DoubleIterator& other) const { return iterator != other.iterator; }
		void operator++()
		{
			++i;
			++iterator;
		}
		auto operator*() const { return std::tie(i, *iterator); }
		size_t i;
		Iterator iterator;
	};
	struct Wrapper
	{
		T& iterable;
		auto begin() { return DoubleIterator{ 0, std::begin(iterable) }; }
		auto end() { return DoubleIterator{ 0, std::end(iterable) }; }
	};
	return Wrapper{ iterable };
}

// This function returns “true” if the two values are roughly equal.
inline bool
approx_equal(double x1, double x2)
{
	return (std::abs(x1 - x2) / ((std::abs(x1) + std::abs(x2)) / 2) < 1e-6);
}

// This function returns “true” if the value is within low and high.
inline bool
within(double low, double value, double high)
{
	return (low <= value) && (value <= high);
}

// Print a message and exit
void error_get_nan_interest_variable(std::str, std::str, msl::RefConst);
void error_get_nan_if_variable(std::str, std::str, msl::RefConst);
void error_get_nan_weight_variable(std::str, std::str, msl::RefConst);
void error_get_negative_weight_variable(std::str, std::str, msl::RefConst);
void error_set_not_alive(std::str, std::str, msl::RefConst);

// Return a error message used by the msl::Exception.
std::str error_as_interest_variable(std::str, std::str);
std::str error_as_if_variable(std::str, std::str);
std::str error_as_weight_variable(std::str, std::str);
std::str error_as_setted_variable(std::str, std::str);

// http://howardhinnant.github.io/date_algorithms.html#days_from_civil
// Returns number of days since civil 1970-01-01.  Negative values indicate days prior to
// 1970-01-01. Preconditions:  y-m-d represents a date in the civil (Gregorian) calendar
//                 m is in [1, 12]
//                 d is in [1, last_day_of_month(y, m)]
//                 y is "approximately" in
//                   [numeric_limits<Int>::min()/366,
//                   numeric_limits<Int>::max()/366]
//                 Exact range of validity is:
//                 [civil_from_days(numeric_limits<Int>::min()),
//                  civil_from_days(numeric_limits<Int>::max()-719468)]
// template <class Int>, Int -> int
[[nodiscard]] int constexpr days_from_civil(int y, unsigned m, unsigned d) noexcept
{
	y -= m <= 2;
	const int era = (y >= 0 ? y : y - 399) / 400;
	const unsigned yoe = static_cast<unsigned>(y - era * 400);           // [0, 399]
	const unsigned doy = (153 * (m + (m > 2 ? -3 : 9)) + 2) / 5 + d - 1; // [0, 365]
	const unsigned doe = yoe * 365 + yoe / 4 - yoe / 100 + doy;          // [0, 146096]
	return era * 146097 + static_cast<int>(doe) - 719468;
}
// Returns year/month/day triple in civil calendar
// Preconditions:  z is number of days since 1970-01-01 and is in the range:
//                   [numeric_limits<Int>::min(), numeric_limits<Int>::max()-719468].
[[nodiscard]] std::tuple<int, unsigned, unsigned> constexpr civil_from_days(int z) noexcept
{
	z += 719468;
	const int era = (z >= 0 ? z : z - 146096) / 146097;
	const unsigned doe = static_cast<unsigned>(z - era * 146097);               // [0, 146096]
	const unsigned yoe = (doe - doe / 1460 + doe / 36524 - doe / 146096) / 365; // [0, 399]
	const int y = static_cast<int>(yoe) + era * 400;
	const unsigned doy = doe - (365 * yoe + yoe / 4 - yoe / 100); // [0, 365]
	const unsigned mp = (5 * doy + 2) / 153;                      // [0, 11]
	const unsigned d = doy - (153 * mp + 2) / 5 + 1;              // [1, 31]
	const unsigned m = mp < 10 ? mp + 3 : mp - 9;                 // [1, 12]
	return { y + (m <= 2), m, d };
}
std::str inline date_to_str(int days)
{
	auto const [y, m, d]{ civil_from_days(days) };
	auto y_{ std::to_string(y) };
	if (y_.size() != 4) {
		if (y < 0) {
			if (y_.size() == 2)
				y_ = std::str{ "-00" } + y_[1];
			else if (y_.size() == 3)
				y_ = std::str{ "-0" } + y_[1] + y_[2];
		} else {
			if (y_.size() == 1)
				y_ = std::str{ "000" } + y_[0];
			else if (y_.size() == 2)
				y_ = std::str{ "00" } + y_[0] + y_[1];
			else if (y_.size() == 3)
				y_ = std::str{ "0" } + y_[0] + y_[1] + y_[2];
		}
	}
	auto const m_{ std::to_string(m) };
	auto const d_{ std::to_string(d) };
	return y_ + '-' + (m_.size() == 2 ? m_ : '0' + m_) + '-' + (d_.size() == 2 ? d_ : '0' + d_);
}

static int to_day{ days_from_civil(0, 1, 1) };

// The definition of the “circular forward iterator”, used to visit each member of a subset of the
// population.
#include "circular-iterator-1.hpp"

// The definition of the “Individual” object.
#include "individual.hpp"

// Uniform syntax for parallelised “Population” function members.
//
// Before:
//          std::count_if(std::execution::par_unseq,
//                      cbegin(population),
//                      cend(population),
//                      [](msl::RefConst ref) { return ref.alive() && ref.family_head; })
//
// After:
//          population.count_if(msl::var::family_head)
//
#include "namespace-var.hpp"

// The definition of the “dump” needs the definition of the “msl::var” namespace as the argument
// by default is the list of all the variables.
void dump(msl::RefConst,
          std::initializer_list<msl::Table> = {
#define CTG(prm, format) msl::var::prm,
#define FLG(prm) msl::var::prm,
#define FLT(prm) msl::var::prm,
#define LNK(prm) msl::var::prm##_head, msl::var::prm##_next, msl::var::prm##_weight,
#define PTR(prm) msl::var::prm,
#include "msl-variables.cpp"
#undef CTG
#undef FLG
#undef FLT
#undef LNK
#undef PTR
          });

// The special getter for “FLT” data member which test for missing value and need the “dump”
// function.
#define CTG(prm, format) /* nothing */
#define FLG(prm)         /* nothing */
#define FLT(prm)                                                                                   \
	inline float msl::Individual::prm##_() const                                                     \
	{                                                                                                \
		if (std::isnan(prm)) {                                                                         \
			std::cerr << "msl library: the value of the variable “" #prm "” is missing for the “"    \
			          << std::to_string(id_number)                                                       \
			          << "” individual when using the getter “msl::Individual::get_" #prm                \
			             "()” member function.\n";                                                       \
			msl::dump(*this);                                                                            \
			std::cerr << "abnormal exit..." << std::endl;                                                \
			exit(1);                                                                                     \
		}                                                                                              \
		return prm;                                                                                    \
	}
#define LNK(prm) /* nothing */
#define PTR(prm) /* nothing */
#include "msl-variables.cpp"
#undef CTG
#undef FLG
#undef FLT
#undef LNK
#undef PTR

// The function member “age” which needs the “dump” function.
inline double
msl::Individual::age() const
{
	if (msl::to_day == days_from_civil(0, 1, 1))
		throw msl::Exception{ "The “to_day” variable has not be setted" };
	if (birthday == days_from_civil(0, 1, 1)) {
		std::cerr << "msl library: The value of the “birthday” variable is missing for the “"
		          << std::to_string(id_number)
		          << "” individual when using the “msl::Population::age” member function.\n";
		dump(*this);
		std::cerr << "Abnormal exit..." << std::endl;
		exit(1);
	};
	if (birthday > msl::to_day) {
		std::cerr << "msl library: The value of the “birthday” variable is greather than the value "
		             "of the “to_day” variable for the “"
		          << std::to_string(id_number)
		          << "” individual when using the “msl::Population::age” member function.\n";
		dump(*this);
		std::cerr << "Abnormal exit..." << std::endl;
		exit(1);
	};
	return (msl::to_day - birthday) / 365.25;
}

#include "circular-iterator-2.hpp"

// Declare all parameters as “extern” values; the parameters are initialised in the “msl.cpp” file
// to Not a Number value ; the function “load_parameters” will set the actual values.
namespace prm {
#define PRM(prm) extern double prm;
#include "msl-parameters.cpp"
#undef PRM
extern std::map<std::str, double*> prm2ptr_;
} // namespace prm

// The value returned when dereferencing some ad hoc iterators.
struct DoubleDouble
{
	double first;
	double second;
};

// The value returned when dereferencing some ad hoc iterator.
struct DoubleUint8
{
	double value;
	uint8_t by;
};

// The “Population” struct is based on the “std::vector<msl::Individual>”.
using PopBase = std::vector<msl::Individual>;
// Different ad hoc iterators are based on the “PopBase::const_iterator”.
using ConstIterator = PopBase::const_iterator;

// The definition of the “Population struct”.
#include "population.hpp"

msl::Population
load_population(std::str const& file_name, std::str const& file_name_2 = std::str{});

msl::Population
transport_load_population(std::str const& file_name);

struct TimeWatch
{
	TimeWatch()
	  : start_(std::chrono::steady_clock::now())
	{
	}
	void start() { start_ = std::chrono::steady_clock::now(); }
	double elapsed() const
	{
		std::chrono::duration<double> diff{ std::chrono::steady_clock::now() - start_ };
		return diff.count();
	}
	std::chrono::time_point<std::chrono::steady_clock> start_;
};

static TimeWatch time_watch{};

} // namespace msl

#endif // MSL_HPP
