double
msl::Population::gini(msl::Lambda const& get, char const* f_name) const
{
	return this->gini_if(get, [](msl::RefConst ref) { return true; });
}
double
msl::Population::gini_if(msl::Lambda const& get,
                         msl::Lambda const& filter,
                         char const* f_name) const
{
	// Get the number of alive individuals in each chunk.
	std::vector<size_t> counts(msl::THREADS_NR);
	auto const task1{ [&](size_t i, size_t begin, size_t end) {
		counts[i] = std::accumulate(
		  this->cbegin() + begin, this->cbegin() + end, size_t{}, [&](size_t count, msl::RefConst ref) {
			  if (ref.alive()) {
				  double const crit{ filter(ref) };
				  // Check if the value of the criterion is not a missing one. Display an error and exit.
				  if (std::isnan(crit))
					  msl::error_get_nan_if_variable(f_name, "<unknow>", ref);
				  // Check if the value is not a missing one. Display an error and exit.
				  if (crit && std::isnan(get(ref)))
					  msl::error_get_nan_interest_variable(f_name, "<unknow>", ref);
				  return crit ? count + 1 : count;
			  }
			  return count;
		  });
	} };
	// Parallelised execution of the task.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	size_t begin{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, this->size()) };
		threads.emplace_back(task1, i, begin, end);
		begin = end;
	}
	msl::join(threads);
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	// We return a missing value to signal the issue.
	if (count == 0)
		return std::numeric_limits<double>::quiet_NaN();
	if (count == 1)
		return 1;
	// Allocate the giant vector of the values to take into account.
	std::vector<float> v(count);
	// The tasks to populate this vector.
	auto const task2{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter(ref))
				v[base++] = get(ref);
		}
	} };
	begin = 0;
	size_t base{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, this->size()) };
		threads[i] = std::thread{ task2, begin, end, base };
		begin = end;
		base += counts[i];
	}
	msl::join(threads);
	// Parallelised sort of the giant vector.
	std::sort(std::execution::par_unseq, v.begin(), v.end());

	// Operator to parallelised compute sum_{i=1}^n i x_i where x_i is a float value from the ''v''
	// vector. Bug: the algorithm ''reduce'' can copy x, invaliding the address of x.
	//	struct SpecialAddReduceOperator
	//	{
	//		SpecialAddReduceOperator(std::vector<float> const& v)
	//		  : base_(v.data())
	//		{}
	//		// Arithmetic on pointer... :(
	//		double operator()(double sum, float const& x) { return sum + (&x - base_ + 1) * x; }
	//		double operator()(float const& x, double sum) { return sum + (&x - base_ + 1) * x; }
	//		double operator()(float const& x1, float const& x2)
	//		{
	//			return (&x1 - base_ + 1) * x1 + (&x2 - base_ + 1) * x2;
	//		}
	//		double operator()(double sum1, double sum2) { return sum1 + sum2; }
	//		float const* base_;
	//	};
	//	// The computation.
	//	auto const sum_i_times_xi{ std::reduce(
	//		std::execution::par_unseq, v.cbegin(), v.cend(), double{}, SpecialAddReduceOperator{ v }) };

	// Compute sum_{i=1}^n i x_i where x_i is a float value from the ''v'' std::vector. Arithmetic on
	// pointer...
	auto const sum_i_times_xi{ std::accumulate(
		v.cbegin(), v.cend(), double{}, [&](double sum, float const& x) {
		  return sum + (&x - v.data() + 1) * x;
		}) };
	// Parallelised computation of sum_{i=1}^n x_i.
	auto const sum{ std::reduce(std::execution::par_unseq, v.cbegin(), v.cend(), double{}) };

	return (2 * sum_i_times_xi / sum - count - 1) / count;
}

// Weighted gini: https://rdrr.io/cran/acid/src/R/weighted.gini.R
// and also the Liam2 sources, with the same reference.

double
msl::Population::weighted_gini(msl::Lambda const& get,
                               msl::Table const& weight,
                               char const* f_name) const
{
	return this->weighted_gini_if(
	  get, [](msl::RefConst ref) { return true; }, weight);
}

double
msl::Population::weighted_gini_if(msl::Lambda const& get,
                                  msl::Lambda const& filter,
                                  msl::Table const& weight,
                                  char const* f_name) const
{
	// Get the number of alive individuals in each chunk.
	std::vector<size_t> counts(msl::THREADS_NR);
	auto const task1{ [&](size_t i, size_t begin, size_t end) {
		counts[i] = std::accumulate(
		  this->cbegin() + begin, this->cbegin() + end, size_t{}, [&](size_t count, msl::RefConst ref) {
			  if (ref.alive()) {
				  double const crit{ filter(ref) };
				  // Check if the value is not a missing one. Display an error and exit.
				  if (std::isnan(crit))
					  msl::error_get_nan_if_variable(f_name, "<unknow>", ref);
				  // Check if the value is not a missing one. Display an error and exit.
				  if (crit) {
					  if (std::isnan(get(ref)))
						  msl::error_get_nan_interest_variable(f_name, "<unknow>", ref);
					  float const w{ weight.flt_(ref) };
					  if (std::isnan(w))
						  msl::error_get_nan_weight_variable(f_name, weight.name_(), ref);
					  if (w < 0)
						  msl::error_get_negative_weight_variable(f_name, weight.name_(), ref);
				  }
				  return crit ? count + 1 : count;
			  }
			  return count;
		  });
	} };
	// Parallelised execution of the task.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	size_t begin{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, this->size()) };
		threads.emplace_back(task1, i, begin, end);
		begin = end;
	}
	msl::join(threads);
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	// We return a missing value to signal the issue.
	if (count == 0)
		return std::numeric_limits<double>::quiet_NaN();
	if (count == 1)
		return 1;
	// Allocate the giant vector of the values to take into account, the first value of the pair
	// is the variable and the second the weight.
	std::vector<std::pair<float, float>> u(count);
	// The tasks to populate this vector.
	auto const task2{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter(ref))
				u[base++] = std::make_pair(get(ref), weight.flt_(ref));
		}
	} };
	begin = 0;
	size_t base{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, this->size()) };
		threads[i] = std::thread{ task2, begin, end, base };
		begin = end;
		base += counts[i];
	}
	msl::join(threads);
	// Parallelised sort of the giant vector.
	std::sort(std::execution::par_unseq, u.begin(), u.end());

	// Switch to ''double'' as ''float'' is not enough accurate. Do not do that before as the sort is
	// faster with ''float''.
	// Cumulated w_i x_i.
	auto cwx{ msl::reserve<double>(count) };
	// Cumulated w_i.
	auto cw{ msl::reserve<double>(count) };
	{
		float const& x{ u[0].first };
		float const& w{ u[0].second };
		cwx.push_back(w * x);
		cw.push_back(w);
	}
	for (size_t i{ 1 }; i < count; ++i) {
		float const& x{ u[i].first };
		float const& w{ u[i].second };
		cwx.push_back(cwx[i - 1] + w * x), cw.push_back(cw[i - 1] + w);
	}

	double sum{};
	for (size_t i{ 1 }; i < count; ++i)
		sum += cw[i - 1] * cwx[i] - cw[i] * cwx[i - 1];

	return sum / (cw.back() * cwx.back());
}
