std::array<double, 101>
msl::Population::pctl(msl::Lambda const& get, char const* f_name) const
{
	return this->pctl_if(
	  get, [](msl::RefConst ref) { return true; }, f_name);
}

std::array<double, 101>
msl::Population::pctl_if(msl::Lambda const& get,
                         msl::Lambda const& filter,
                         char const* f_name) const
{
	// Get the number of alive individuals in each chunk.
	std::vector<size_t> counts(msl::THREADS_NR);
	auto const task1{ [&](size_t i, size_t begin, size_t end) {
		counts[i] = std::accumulate(
		  this->cbegin() + begin, this->cbegin() + end, size_t{}, [&](size_t count, msl::RefConst ref) {
			  if (ref.alive()) {
				  double const crit{ filter(ref) };
				  // Check if the value of the criterion is not a missing one. Display an error and exit.
				  if (std::isnan(crit))
					  msl::error_get_nan_if_variable(f_name, "<unknow>", ref);
				  // Check if the value is not a missing one. Display an error and exit.
				  if (crit && std::isnan(get(ref)))
					  msl::error_get_nan_interest_variable(f_name, "<unknow>", ref);
				  return crit ? count + 1 : count;
			  }
			  return count;
		  });
	} };
	// Parallelised execution of the task.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	size_t begin{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, this->size()) };
		threads.emplace_back(task1, i, begin, end);
		begin = end;
	}
	msl::join(threads);
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::array<double, 101> rvo;
	// We return a array of missing values to signal the issue.
	if (count == 0) {
		rvo.fill(std::numeric_limits<double>::quiet_NaN());
		return rvo;
	}
	// Allocate the giant vector of the values to take into account.
	std::vector<float> v(count);
	// The tasks to populate this vector.
	auto const task2{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter(ref))
				v[base++] = get(ref);
		}
	} };
	begin = 0;
	size_t base{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, this->size()) };
		threads[i] = std::thread{ task2, begin, end, base };
		begin = end;
		base += counts[i];
	}
	msl::join(threads);
	// Parallelised sort of the giant vector.
	std::sort(std::execution::par_unseq, v.begin(), v.end());
	// Compute the percentiles.
	rvo[0] = v.front();
	rvo[100] = v.back();
	// If size = 2, the median index is .5 = 50 / 100. * 2 = .5 * 2 - .5, i.e. i / 100. * v.size() -
	// .5. The index is zero based, then the - .5.
	for (size_t i{ 1 }; i < 100; ++i) {
		double const y{ i / 100. * v.size() - .5 };
		size_t const left_index{ y < 0 ? size_t{} : size_t(y) };
		size_t const right_index{ left_index < v.size() - 1 ? left_index + 1 : v.size() - 1 };
		double const lambda{ y - left_index };
		auto const& x1{ v[left_index] };
		auto const& x2{ v[right_index] };
		rvo[i] = x1 + lambda * (x2 - x1);
	}
	return rvo;
}
std::array<double, 101>
msl::Population::weighted_pctl(msl::Lambda const& get,
                               msl::Table const& weight,
                               char const* f_name) const
{
	return this->weighted_pctl_if(
	  get, [](msl::RefConst ref) { return true; }, weight);
}
std::array<double, 101>
msl::Population::weighted_pctl_if(msl::Lambda const& get,
                                  msl::Lambda const& filter,
                                  msl::Table const& weight,
                                  char const* f_name) const
{
	if (weight.flt_ == nullptr)
		throw msl::Exception{ error_as_weight_variable(f_name, weight.name_()) };
	// Get the number of alive individuals in each chunk.
	std::vector<size_t> counts(msl::THREADS_NR);
	auto const task1{ [&](size_t i, size_t begin, size_t end) {
		counts[i] = std::accumulate(
		  this->cbegin() + begin, this->cbegin() + end, size_t{}, [&](size_t count, msl::RefConst ref) {
			  if (ref.alive()) {
				  double const crit{ filter(ref) };
				  // Check if the value is not a missing one. Display an error and exit.
				  if (std::isnan(crit))
					  msl::error_get_nan_if_variable(f_name, "<unknow>", ref);
				  // Check if the value is not a missing one. Display an error and exit.
				  if (crit) {
					  if (std::isnan(get(ref)))
						  msl::error_get_nan_interest_variable(f_name, "<unknow>", ref);
					  float const w{ weight.flt_(ref) };
					  if (std::isnan(w))
						  msl::error_get_nan_weight_variable(f_name, weight.name_(), ref);
					  if (w < 0)
						  msl::error_get_negative_weight_variable(f_name, weight.name_(), ref);
				  }
				  return crit ? count + 1 : count;
			  }
			  return count;
		  });
	} };
	// Parallelised execution of the task.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	size_t begin{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, this->size()) };
		threads.emplace_back(task1, i, begin, end);
		begin = end;
	}
	msl::join(threads);
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::array<double, 101> rvo;
	// We return a array of missing values to signal the issue.
	if (count == 0) {
		rvo.fill(std::numeric_limits<double>::quiet_NaN());
		return rvo;
	}
	// Allocate the giant vector of the values to take into account.
	std::vector<std::pair<float, float>> v(count);
	// The tasks to populate this vector.
	auto const task2{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter(ref))
				v[base++] = std::make_pair(get(ref), weight.flt_(ref));
		}
	} };
	begin = 0;
	size_t base{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, this->size()) };
		threads[i] = std::thread{ task2, begin, end, base };
		begin = end;
		base += counts[i];
	}
	msl::join(threads);
	// Parallelised sort of the giant vector.
	std::sort(std::execution::par_unseq, v.begin(), v.end());
	// Partial sum of the weights.
	for (size_t i{ 1 }; i < count; ++i)
		v[i].second += v[i - 1].second;

	// Compute the percentiles...
	rvo[0] = v.front().first;
	rvo[100] = v.back().first;
	auto const& total_weight{ v.back().second };
	size_t right_index{};

	for (size_t i{ 1 }; i < 100; ++i) {
		// If the total_weight == count, we get the same expression than in the unweigthed case with
		// a 1 based index.
		auto const y{ (i / 100. + .5 / count) * total_weight };
		while ((v[right_index].second <= y) && (right_index < v.size()))
			++right_index;
		auto const left_index{ right_index > 1 ? right_index - 1 : size_t{} };
		auto const& x1{ v[left_index].first };
		auto const& x2{ v[right_index].first };
		auto const& y1{ v[left_index].second };
		auto const& y2{ v[right_index].second };
		if (y1 == y2)
			rvo[i] = x1;
		else {
			auto const lambda{ (y - y1) / (y2 - y1) };
			rvo[i] = x1 + lambda * (x2 - x1);
		}
	}
	return rvo;
}
