// File to build a static library.
#include "msl.hpp"

// To include the fd-read-xlsx and the fd-write-xlsx stuff in the library.
#include "fd-read-xlsx-header-only.hpp"
#include "fd-write-xlsx-header-only.hpp"

namespace msl {

// The mutex to ensure exclusive use of the data.
std::mutex global_lock;

// Define the formats map and initialise it to a semi-empty map; the “load_formats” function will
// set the formats maps to the actual value.
std::map<msl::Ctg, msl::FormatMap> global_formats{
#define LNK(prm) /* nothing */
#define FLT(prm) /* nothing */
#define FLG(prm) /* nothing */
// Save the format name as the first element of the pair.
#define CTG(prm, format)                                                                           \
	{ msl::var::prm.ctg_, std::pair{ #format, std::map<uint8_t, std::str>{} } },
#define PTR(prm) /* nothing */
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
};

// Define the parameters and initialise them to Not A Number; the “load_parameters” function will
// set the parameters to the actual value. Declare also the “prm2ptr_” map and initialize it.
namespace prm {
#define PRM(prm) double prm{ std::numeric_limits<double>::quiet_NaN() };
#include "msl-parameters.cpp"
#undef PRM
std::map<std::str, double*> prm2ptr_{
#define PRM(prm) { #prm, &prm },
#include "msl-parameters.cpp"
#undef PRM
};
} // namespace prm

void
dump(msl::RefConst ref, std::initializer_list<msl::Table> tables)
{
	std::cerr << "Dump for individual id number “" << ref.id_number << "”, alive{"
	          << (ref.alive() ? "Yes" : "No") << "}, birthday{"
	          << (ref.birthday == msl::days_from_civil(0, 1, 1) ? "." : date_to_str(ref.birthday))
	          << "}.\n";
	bool start{ true };
	for (auto const& table : tables) {
		if (start)
			start = false;
		else
			std::cerr << ", ";
		std::cerr << table.name_() << '{' << table.str_(ref) << '}';
	}
	std::cerr << ".\n";
}

void
error_get_nan_interest_variable(std::str f_name, std::str v_name, msl::RefConst ref)
{
	std::cerr << "msl library: the value of the variable of interest “" << v_name
	          << "” is missing for the “" << std::to_string(ref.id_number)
	          << "” individual when using as a argument for the “msl::Population::" << f_name
	          << "” member function.\n";
	dump(ref);
	std::cerr << "abnormal exit..." << std::endl;
	exit(1);
};
void
error_get_nan_if_variable(std::str f_name, std::str v_name, msl::RefConst ref)
{
	std::cerr << "msl library: The value of the “if” variable “" << v_name << "” is missing for the “"
	          << std::to_string(ref.id_number)
	          << "” individual when using as a argument for the “msl::Population::" << f_name
	          << "” member function.\n";
	dump(ref);
	std::cerr << "Abnormal exit..." << std::endl;
	exit(1);
};
void
error_get_nan_weight_variable(std::str f_name, std::str v_name, msl::RefConst ref)
{
	std::cerr << "msl library: The value of the variable of weight “" << v_name
	          << "” is missing for the “" << std::to_string(ref.id_number)
	          << "” individual when using as a argument for the “msl::Population::" << f_name
	          << "” member function.\n";
	dump(ref);
	std::cerr << "Abnormal exit..." << std::endl;
	exit(1);
};
void
error_get_negative_weight_variable(std::str f_name, std::str v_name, msl::RefConst ref)
{
	std::cerr << "msl library: The value of the variable of weight “" << v_name
	          << "” is negative for the “" << std::to_string(ref.id_number)
	          << "” individual when using as a argument for the “msl::Population::" << f_name
	          << "” member function.\n";
	dump(ref);
	std::cerr << "Abnormal exit..." << std::endl;
	exit(1);
};
void
error_set_not_alive(std::str f_name, std::str v_name, msl::RefConst ref)
{
	std::cerr << "msl library: The individual “" << std::to_string(ref.id_number)
	          << "” is not an alive one when trying to set the variable “" << v_name
	          << "” using the “msl::Population::" << f_name << "” member function.\n";
	dump(ref);
	std::cerr << "Abnormal exit..." << std::endl;
	exit(1);
};
std::str
error_as_interest_variable(std::str f_name, std::str v_name)
{
	return "Invalid use of the “" + v_name +
	       "” variable as variable of interest for the “Population::" + f_name + "” function member.";
}
std::str
error_as_if_variable(std::str f_name, std::str v_name)
{
	return "Invalid use of the “" + v_name +
	       "” variable as “if” variable for the “Population::" + f_name + "” function member.";
}
std::str
error_as_weight_variable(std::str f_name, std::str v_name)
{
	return "Invalid use of the “" + v_name +
	       "” variable as “weight” variable for the “Population::" + f_name + "” function member.";
}
std::str
error_as_by_variable(std::str f_name, std::str v_name)
{
	return "Invalid use of the “" + v_name +
	       "” variable as “by” variable for the “Population::" + f_name + "” function member.";
}
std::str
error_as_setted_variable(std::str f_name, std::str v_name)
{
	return "Invalid use of the “" + v_name +
	       "” variable as a variable to set for the “Population::" + f_name + "” function member.";
}

//====================//
// For each functions //
//====================//

// Parallelised “for_each” only over alive individuals:
//       population.for_each([](msl::Ref ref){ some_function(ref); })
void
msl::Population::for_each(msl::Function const& function)
{
	std::for_each(std::execution::par_unseq, this->begin(), this->end(), [&](msl::Ref ref) {
		// Call the function only on alive individuals.
		if (ref.alive())
			function(ref);
	});
}

// Parallelised and filtered “for_each” only over alive individuals:
//       population.for_each([](msl::Ref ref){ some_function(ref); },
//                           [](msl::RefConst ref){ return ref.disposable_income > 0; })
void
msl::Population::for_each_if(msl::Function const& function, msl::Lambda const& filter)
{
	std::for_each(std::execution::par_unseq, this->begin(), this->end(), [&](msl::Ref ref) {
		// Call the function only on alive individuals.
		if (ref.alive()) {
			double const crit{ filter(ref) };
			// Check if the value is not a missing one. Display an error and exit.
			if (std::isnan(crit))
				error_get_nan_if_variable("for_each_if", "<lambda>", ref);
			// Call the function on this individual.
			if (crit)
				function(ref);
		}
	});
}

// Parallelised and filtered “for_each” only over alive individuals:
//       population.for_each([](msl::Ref ref){ some_function(ref); }, msl::var::family_head)
void
msl::Population::for_each_if(msl::Function const& function, msl::Table const& filter)
{
	if (filter.flt_ == nullptr)
		throw msl::Exception{ error_as_if_variable("for_each_if", filter.name_()) };

	std::for_each(std::execution::par_unseq, this->begin(), this->end(), [&](msl::Ref ref) {
		// Call the function only on alive individuals.
		if (ref.alive()) {
			float const crit{ filter.flt_(ref) };
			// Check if the value is not a missing one. Display an error and exit.
			if (std::isnan(crit))
				error_get_nan_if_variable("for_each_if", filter.name_(), ref);
			// Call the function on this individual.
			if (crit)
				function(ref);
		}
	});
}

//====================//
// Count if functions //
//====================//

// Parallelised “count_if” only over the alive individuals:
//       population.count_if(msl::var::female)
size_t
msl::Population::count_if(msl::Lambda const& filter) const
{
	return std::count_if(
	  std::execution::par_unseq, this->cbegin(), this->cend(), [&](msl::RefConst ref) {
		  // If the individual is not a alive one, return 0 as this value is interpreted as “false”.
		  if (!ref.alive())
			  return double{};
		  double const crit{ filter(ref) };
		  // Check if the value is not a missing one. Display an error and exit.
		  if (std::isnan(crit))
			  msl::error_get_nan_if_variable("<lambda>", "count_if", ref);
		  return crit;
	  });
}

// Parallelised “count_if” only over the alive individuals:
//       population.count_if([](msl::RefConst ref){ return ref.disposable_income > 0; })
size_t
msl::Population::count_if(msl::Table const& filter) const
{
	if (filter.flt_ == nullptr)
		throw msl::Exception{ error_as_if_variable("count_if", filter.name_()) };
	return std::count_if(
	  std::execution::par_unseq, this->cbegin(), this->cend(), [&](msl::RefConst ref) {
		  // If the individual is not a alive one, return 0 as this value is interpreted as “false”.
		  if (!ref.alive())
			  return float{};
		  float const crit{ filter.flt_(ref) };
		  // Check if the value is not a missing one. Display an error and exit.
		  if (std::isnan(crit))
			  msl::error_get_nan_if_variable(filter.name_(), "count_if", ref);
		  return crit;
	  });
}

//===============//
// Sum functions //
//===============//

#include "include-sum.cpp"

//===============//
// Avg functions //
//===============//

#include "include-avg.cpp"

//===============//
// Std functions //
//===============//

#include "include-std.cpp"

//===============//
// Min functions //
//===============//

#include "include-min.cpp"

//===============//
// Max functions //
//===============//

#include "include-max.cpp"

//=======================//
// Percentiles functions //
//=======================//

#include "include-pctl.cpp"

//================//
// Gini functions //
//================//

#include "include-gini.cpp"

//==================//
// Sum by functions //
//==================//

#include "include-sum-by.cpp"

namespace internals {
std::vector<size_t>
counts_by_chunk(msl::Population const& population, std::vector<std::thread>& threads, size_t step)
{
	// Get the number of alive individuals in each chunk.
	std::vector<size_t> rvo(msl::THREADS_NR);
	auto const task{ [&](size_t i, size_t begin, size_t end) {
		rvo[i] = std::accumulate(
		  population.cbegin() + begin,
		  population.cbegin() + end,
		  size_t{},
		  [&](size_t count, msl::RefConst ref) { return ref.alive() ? count + 1 : count; });
	} };
	// Parallelised execution of the task.
	size_t begin{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, population.size()) };
		threads.emplace_back(task, i, begin, end);
		begin = end;
	}
	msl::join(threads);
	return rvo;
}
void
fill_rvo(std::vector<size_t>& rvo,
         std::function<void(size_t, size_t, size_t)> const& task,
         Population const& population,
         std::vector<std::thread>& threads,
         size_t step,
         std::vector<size_t> const& counts)
{
	size_t begin{};
	size_t base{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, population.size()) };
		threads[i] = std::thread{ task, begin, end, base };
		begin = end;
		base += counts[i];
	}
	msl::join(threads);
}
void
fill_rvo(std::vector<double>& rvo,
         std::function<void(size_t, size_t, size_t)> const& task,
         Population const& population,
         std::vector<std::thread>& threads,
         size_t step,
         std::vector<size_t> const& counts)
{
	size_t begin{};
	size_t base{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, population.size()) };
		threads[i] = std::thread{ task, begin, end, base };
		begin = end;
		base += counts[i];
	}
	msl::join(threads);
}
} // namespace internals

std::vector<size_t>
msl::Population::get_indexes(char const* f_name) const
{
	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	// The total range is splitted in chunks with step width.
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	// Number of alive individuals in each chunk.
	auto const counts{ msl::internals::counts_by_chunk(*this, threads, step) };
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	// The vector of indexes to return.
	std::vector<size_t> rvo(count);
	if (count == 0)
		return rvo;
	// The tasks to populate this vector.
	auto const task{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive())
				rvo[base++] = i;
		}
	} };
	// Parallelised fill of the vector.
	msl::internals::fill_rvo(rvo, task, *this, threads, step, counts);
	return rvo;
}
std::vector<double>
msl::Population::get_variable(msl::Lambda const& get, char const* f_name) const
{

	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	auto const counts{ msl::internals::counts_by_chunk(*this, threads, step) };
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::vector<double> rvo(count);
	if (count == 0)
		return rvo;
	// The tasks to populate this vector.
	auto const task{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive()) {
				double const x{ get(ref) };
				// Check if the value is not a missing one. Display an error and exit.
				if (std::isnan(x))
					msl::error_get_nan_interest_variable(f_name, "<lambda>", ref);
				rvo[base++] = x;
			}
		}
	} };
	msl::internals::fill_rvo(rvo, task, *this, threads, step, counts);
	return rvo;
}
std::vector<double>
msl::Population::get_variable(msl::Table const& get, char const* f_name) const
{
	if (get.flt_ == nullptr)
		throw msl::Exception{ error_as_interest_variable(f_name, get.name_()) };
	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	auto const counts{ msl::internals::counts_by_chunk(*this, threads, step) };
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::vector<double> rvo(count);
	if (count == 0)
		return rvo;
	// The tasks to populate this vector.
	auto const task{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive()) {
				float const x{ get.flt_(ref) };
				// Check if the value is not a missing one. Display an error and exit.
				if (std::isnan(x))
					msl::error_get_nan_interest_variable(f_name, get.name_(), ref);
				rvo[base++] = x;
			}
		}
	} };
	msl::internals::fill_rvo(rvo, task, *this, threads, step, counts);
	return rvo;
}

namespace internals {
std::vector<size_t>
counts_by_chunk_if(msl::Population const& population,
                   std::vector<std::thread>& threads,
                   size_t step,
                   msl::Lambda const& filter,
                   char const* f_name)
{
	// Get the number of alive individuals in each chunk.
	std::vector<size_t> rvo(msl::THREADS_NR);
	auto const task{ [&](size_t i, size_t begin, size_t end) {
		rvo[i] = std::accumulate(population.cbegin() + begin,
		                         population.cbegin() + end,
		                         size_t{},
		                         [&](size_t count, msl::RefConst ref) {
			                         if (ref.alive()) {
				                         double const crit{ filter(ref) };
				                         // Check if the value is not a missing one. Display an error and
				                         // exit.
				                         if (std::isnan(crit))
					                         msl::error_get_nan_if_variable(f_name, "<lambda>", ref);
				                         return crit ? count + 1 : count;
			                         }
			                         return count;
		                         });
	} };
	// Parallelised execution of the task.
	size_t begin{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, population.size()) };
		threads.emplace_back(task, i, begin, end);
		begin = end;
	}
	msl::join(threads);
	return rvo;
}
} // namespace internals

std::vector<size_t>
msl::Population::get_indexes_if(msl::Lambda const& filter, char const* f_name) const
{
	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	auto const counts{ msl::internals::counts_by_chunk_if(*this, threads, step, filter, f_name) };
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::vector<size_t> rvo(count);
	if (count == 0)
		return rvo;
	// The tasks to populate this vector.
	auto const task{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter(ref))
				rvo[base++] = i;
		}
	} };
	msl::internals::fill_rvo(rvo, task, *this, threads, step, counts);
	return rvo;
}
namespace internals {
std::vector<size_t>
counts_by_chunk_if(msl::Population const& population,
                   std::vector<std::thread>& threads,
                   size_t step,
                   msl::Table const& filter,
                   char const* f_name)
{
	// Get the number of alive individuals in each chunk.
	std::vector<size_t> rvo(msl::THREADS_NR);
	auto const task{ [&](size_t i, size_t begin, size_t end) {
		rvo[i] = std::accumulate(population.cbegin() + begin,
		                         population.cbegin() + end,
		                         size_t{},
		                         [&](size_t count, msl::RefConst ref) {
			                         if (ref.alive()) {
				                         float const crit{ filter.flt_(ref) };
				                         // Check if the value is not a missing one. Display an error and
				                         // exit.
				                         if (std::isnan(crit))
					                         msl::error_get_nan_if_variable(f_name, filter.name_(), ref);
				                         return crit ? count + 1 : count;
			                         }
			                         return count;
		                         });
	} };
	// Parallelised execution of the task.
	size_t begin{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, population.size()) };
		threads.emplace_back(task, i, begin, end);
		begin = end;
	}
	msl::join(threads);
	return rvo;
}
} // namespace internals

std::vector<size_t>
msl::Population::get_indexes_if(msl::Table const& filter, char const* f_name) const
{
	if (filter.flt_ == nullptr)
		throw msl::Exception{ error_as_if_variable(f_name, filter.name_()) };
	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	auto const counts{ msl::internals::counts_by_chunk_if(*this, threads, step, filter, f_name) };
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::vector<size_t> rvo(count);
	if (count == 0)
		return rvo;
	// The tasks to populate this vector.
	auto const task{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter.flt_(ref))
				rvo[base++] = i;
		}
	} };
	msl::internals::fill_rvo(rvo, task, *this, threads, step, counts);
	return rvo;
}
std::vector<double>
msl::Population::get_variable_if(msl::Lambda const& get,
                                 msl::Lambda const& filter,
                                 char const* f_name) const
{
	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	auto const counts{ msl::internals::counts_by_chunk_if(*this, threads, step, filter, f_name) };
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::vector<double> rvo(count);
	if (count == 0)
		return rvo;
	// The tasks to populate this vector.
	auto const task{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter(ref)) {
				double const x{ get(ref) };
				// Check if the value is not a missing one. Display an error and exit.
				if (std::isnan(x))
					msl::error_get_nan_interest_variable(f_name, "<lambda>", ref);
				rvo[base++] = x;
			}
		}
	} };
	msl::internals::fill_rvo(rvo, task, *this, threads, step, counts);
	return rvo;
}
std::vector<double>
msl::Population::get_variable_if(msl::Lambda const& get,
                                 msl::Table const& filter,
                                 char const* f_name) const
{
	if (filter.flt_ == nullptr)
		throw msl::Exception{ error_as_if_variable(f_name, filter.name_()) };
	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	auto const counts{ msl::internals::counts_by_chunk_if(*this, threads, step, filter, f_name) };
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::vector<double> rvo(count);
	if (count == 0)
		return rvo;
	// The tasks to populate this vector.
	auto const task{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter.flt_(ref)) {
				double const x{ get(ref) };
				// Check if the value is not a missing one. Display an error and exit.
				if (std::isnan(x))
					msl::error_get_nan_interest_variable(f_name, "<lambda>", ref);
				rvo[base++] = x;
			}
		}
	} };
	msl::internals::fill_rvo(rvo, task, *this, threads, step, counts);
	return rvo;
}
std::vector<double>
msl::Population::get_variable_if(msl::Table const& get,
                                 msl::Lambda const& filter,
                                 char const* f_name) const
{
	if (get.flt_ == nullptr)
		throw msl::Exception{ error_as_interest_variable(f_name, get.name_()) };
	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	auto const counts{ msl::internals::counts_by_chunk_if(*this, threads, step, filter, f_name) };
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::vector<double> rvo(count);
	if (count == 0)
		return rvo;
	// The tasks to populate this vector.
	auto const task{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter(ref)) {
				float const x{ get.flt_(ref) };
				// Check if the value is not a missing one. Display an error and exit.
				if (std::isnan(x))
					msl::error_get_nan_interest_variable(f_name, get.name_(), ref);
				rvo[base++] = x;
			}
		}
	} };
	msl::internals::fill_rvo(rvo, task, *this, threads, step, counts);
	return rvo;
}
std::vector<double>
msl::Population::get_variable_if(msl::Table const& get,
                                 msl::Table const& filter,
                                 char const* f_name) const
{
	if (get.flt_ == nullptr)
		throw msl::Exception{ error_as_interest_variable(f_name, get.name_()) };
	if (filter.flt_ == nullptr)
		throw msl::Exception{ error_as_if_variable(f_name, filter.name_()) };
	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	auto const counts{ msl::internals::counts_by_chunk_if(*this, threads, step, filter, f_name) };
	// The total number of values to take into account.
	auto const count{ std::accumulate(counts.begin(), counts.end(), size_t{}) };
	std::vector<double> rvo(count);
	if (count == 0)
		return rvo;
	// The tasks to populate this vector.
	auto const task{ [&](size_t begin, size_t end, size_t base) {
		for (size_t i{ begin }; i < end; ++i) {
			msl::RefConst ref{ *(this->cbegin() + i) };
			if (ref.alive() && filter.flt_(ref)) {
				float const x{ get.flt_(ref) };
				// Check if the value is not a missing one. Display an error and exit.
				if (std::isnan(x))
					msl::error_get_nan_interest_variable(f_name, get.name_(), ref);
				rvo[base++] = x;
			}
		}
	} };
	msl::internals::fill_rvo(rvo, task, *this, threads, step, counts);
	return rvo;
}
void
msl::Population::set_variable(msl::Table const& set,
                              std::vector<size_t> const& indexes,
                              std::vector<double> const& values,
                              char const* f_name)
{
	if (set.set_ == nullptr)
		throw msl::Exception{ error_as_setted_variable(f_name, set.name_()) };
	if (indexes.size() != values.size())
		throw msl::Exception{ std::str{
			                      "Sizes of indexes and values vector are not the same in the “" } +
			                    f_name + "” function member" };
	auto const task{ [&](size_t begin, size_t end) {
		for (size_t i{ begin }; i < end; ++i) {
			auto const ii{ indexes[i] };
			if (!(*this)[ii].alive())
				error_set_not_alive(f_name, set.name_(), (*this)[ii]);
			// Narrowing...
			set.set_((*this)[ii], values[i]);
		}
	} };
	// A thread for each core.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (indexes.size() - 1) / msl::THREADS_NR + 1 };
	size_t begin{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, indexes.size()) };
		threads.emplace_back(task, begin, end);
		begin = end;
	}
	msl::join(threads);
}

// Find a free slot for a new individual. We try to keep balanced the free slots within the
// population so the burden of distributed computations will be also balanced.
msl::Ref
msl::Population::birth(msl::Ref parent, int birthday, size_t id_number)
{
	if (!parent.alive())
		throw msl::Exception{
			"the parent is not an alive individual in the Population::birth() member function"
		};
	std::lock_guard<std::mutex> lock{ msl::global_lock };
	// If the size of population is 1000 and THREADS_NR is equal to 12, we consider the thunks
	//    [0, 84[, [84, 168[, ... [924, 1000[
	// Within the first chunk, we do a parallel find. If we get a free slot, it is ok. If not, we try
	// the second chunk, etc. The first chunk is circularly selected from the call number. It is not
	// necessary a good idea as the new individuals are dispersed throughout the population.
	static size_t call_nr{};
	static size_t next_id_number{};
	auto const first_range{ (this->size() - 1) / msl::THREADS_NR + 1 };
	for (size_t j{}; j < msl::THREADS_NR; ++j) {
		auto begin{ call_nr * first_range };
		auto end{ std::min(begin + first_range, this->size()) };
		call_nr = (call_nr + 1) % (msl::THREADS_NR + 1);
		auto const iterator{ std::find_if(
			std::execution::par, this->begin() + begin, this->begin() + end, [](msl::Ref ref) {
			  return !ref.alive();
			}) };
		if (iterator != this->begin() + end) {
			if (iterator->alive()) {
				std::cerr << "msl library message: Internal error: “msl::Population::birth“" << std::endl;
				exit(1);
			}
			iterator->set_alive(true).set_birthday(birthday);
			// The user do not give a individual number.
			if (id_number == 0) {
				// At previous call, the user do not give a individual number. Use the stored one.
				if (next_id_number != 0) {
					iterator->id_number = next_id_number;
					// Store a incremented number to, possibly, use it next call.
					++next_id_number;
				}
				// Find the largest individual number used in the data and fix the individual number to max
				// + 1. Missing values are less than other values.
				else {
					auto const n =
					  std::max_element(std::execution::par,
					                   this->cbegin(),
					                   this->cend(),
					                   [](msl::RefConst ref1, msl::RefConst ref2) {
						                   if (ref1.alive())
							                   return ref2.alive() ? ref1.id_number < ref2.id_number : false;
						                   else
							                   return ref2.alive();
					                   })
					    ->id_number +
					  1;
					iterator->id_number = n;
					// Store a incremented number to, possibly, use it next call.
					next_id_number = n + 1;
				}
			} else {
				// Use the number provided by the user and invalidate the stored number.
				iterator->id_number = id_number;
				next_id_number = 0;
			}
			// First, put the individual in a valid state: the individual is the unique member of all the
			// subsets defined in the model. Second, put the individual in the parent subset. Use the
			// member function without lock as a lock is already set.
#define LNK(prm)                                                                                   \
	iterator->set_##prm##_head(true);                                                                \
	iterator->prm##_next = &*iterator;                                                               \
	parent.merge_without_lock_##prm(*iterator);
#define FLT(prm)         /* nothing */
#define FLG(prm)         /* nothing */
#define CTG(prm, format) /* nothing */
#define PTR(prm)         /* nothing */
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
			// Return a reference on the new individual.
			return *iterator;
		}
	}
	throw msl::Exception{ "unable to find a free slot in the Population::birth() member function" };
}

msl::RefConst
msl::Population::find(size_t id_number) const
{
	auto const iterator{ std::find_if(
		std::execution::par_unseq, this->cbegin(), this->cend(), [&](msl::RefConst ref) {
		  return ref.alive() && ref.id_number == id_number;
		}) };
	if (iterator == this->cend())
		throw msl::Exception("Number not found");
	else
		return *iterator;
}

msl::Ref
msl::Population::find(size_t id_number)
{
	auto const iterator{ std::find_if(
		std::execution::par_unseq, this->begin(), this->end(), [&](msl::RefConst ref) {
		  return ref.alive() && ref.id_number == id_number;
		}) };
	if (iterator == this->cend())
		throw msl::Exception("Number not found");
	else
		return *iterator;
}

// Simulate the event (by calling the “function” function on the selected individuals) with the
// probability returned by the “get” function over the
size_t
msl::Population::align(size_t n, msl::Lambda const& get, msl::Function const& function, bool debug)
{
	// Sum of the probabilities within each chunk.
	std::vector<double> proba_sums(msl::THREADS_NR, 0);
	// Number of eligible individuals within each chunk.
	std::vector<size_t> eligibles(msl::THREADS_NR, 0);
	// Fist task: compute, for each chunks (then each thread), the sum of the probabilities and the
	// number of eligible individuals.
	auto const task1{ [&](size_t i, size_t begin, size_t end) {
		std::pair<double, size_t> pair{ std::accumulate(
			this->cbegin() + begin,
			this->cbegin() + end,
			std::pair<double, size_t>{},
			[&](std::pair<double, size_t> pair, msl::RefConst ref) {
			  if (ref.alive()) {
				  auto const proba{ get(ref) };
				  if (proba < 0) {
					  auto const message{
						  "msl library message: Population::align: probability < 0 for the individual " +
						  std::to_string(ref.id_number) + ".\n"
					  };
					  std::cerr << message;
					  std::cerr.flush();
					  return std::pair<double, size_t>{ std::numeric_limits<double>::quiet_NaN(), 0 };
				  } else if (proba > 1) {
					  auto const message{
						  "msl library message: Population::align: probability > 1 for the individual " +
						  std::to_string(ref.id_number) + ".\n"
					  };
					  std::cerr << message;
					  std::cerr.flush();
					  return std::pair<double, size_t>{ std::numeric_limits<double>::quiet_NaN(), 0 };
				  } else if (std::isnan(proba)) {
					  auto const message{ "msl library message: Population::align: probability missing value "
						                    "for the individual " +
						                    std::to_string(ref.id_number) + ".\n" };
					  std::cerr << message;
					  std::cerr.flush();
					  return std::pair<double, size_t>{ std::numeric_limits<double>::quiet_NaN(), 0 };
				  } else
					  return std::pair<double, size_t>{ pair.first + proba, pair.second + (proba > 0) };
			  } else
				  return pair;
			}) };
		proba_sums[i] = pair.first;
		eligibles[i] = pair.second;
	} };

	// Parallelised execution of the task.
	auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
	auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
	size_t begin{};
	for (size_t i{}; i < msl::THREADS_NR; ++i) {
		auto const end{ std::min(begin + step, this->size()) };
		threads.emplace_back(task1, i, begin, end);
		begin = end;
	}
	msl::join(threads);

	// Compute the agregate number of events as the sum of the probabilities.
	double const sum_proba{ std::accumulate(proba_sums.cbegin(), proba_sums.cend(), double{}) };
	if (std::isnan(sum_proba)) {
		std::cerr << "msl library message: Population::align: premature exit.\n";
		return 0;
	}
	// Compute the number of eligible individuals.
	double const eligible_nr{ std::accumulate(eligibles.cbegin(), eligibles.cend(), double{}) };
	if (eligible_nr < n) {
		std::cerr << "msl library message: Population::align: not enought eligible individuals.\n";
		return 0;
	}
	auto events_nr{ proba_sums };
	// Alignement correction factor: each individual probability must be multiplied by this factor.
	auto const factor{ n / sum_proba };
	// Set the number of events within each chunks with respect to the alignement.
	for (size_t i{ 1 }; i < msl::THREADS_NR; ++i)
		events_nr[i] = std::min(size_t(factor * proba_sums[i] + .5), eligibles[i]);
	// To meet exactly the alignement constraint, the number of the thread #0 is computed by
	// substraction.
	events_nr[0] = n - std::accumulate(events_nr.cbegin() + 1, events_nr.cend(), double{});
	// Oups, not enough individuals in thread #0... Dispatch the excess on the other threads.
	if (events_nr[0] > eligibles[0]) {
		auto remainder{ events_nr[0] - eligibles[0] };
		events_nr[0] = eligibles[0];
		for (size_t i{ 1 }; i < msl::THREADS_NR; ++i) {
			if (events_nr[i] < eligibles[i]) {
				auto const nr{ std::max(remainder, eligibles[i] - events_nr[i]) };
				events_nr[i] += nr;
				remainder -= nr;
				if (remainder == 0)
					break;
			}
		}
	}
	// Oups... Dispatch the shortage on the others threads.
	else if (events_nr[0] < 0) {
		auto remainder{ -events_nr[0] };
		events_nr[0] = 0;
		for (size_t i{ 1 }; i < msl::THREADS_NR; ++i) {
			if (events_nr[i] > 0) {
				auto const nr{ std::min(remainder, events_nr[i]) };
				events_nr[i] -= nr;
				remainder -= nr;
				if (remainder == 0)
					break;
			}
		}
	}

	// Step 2: selection of the individuals within each chunk. Do not execute the user function as
	// this function can modify the population.
	std::vector<size_t> sims(msl::THREADS_NR, 0);
	std::vector<std::vector<size_t>> indexes(msl::THREADS_NR);
	for (auto [i, ref] : msl::enumerate(indexes))
		ref.reserve(events_nr[i]);

	// Second task: align the events within each chunk.
	auto const task2{ [&](size_t i, size_t begin, size_t end) {
		// No events to simulate…
		if (events_nr[i] == 0)
			return;
		std::mt19937 generator;
		// Thread specific initialisation of the pseudo number generator.
		generator.seed(i);
		// Adaptator to get a distribution between 0 and 1.
		std::uniform_real_distribution<> unif{ 0, 1 };
		auto remainder{ events_nr[i] };
		auto eligible_nr{ eligibles[i] };
		size_t n{};
		size_t count{ end - begin };
		for (size_t j{}; j < count; ++j) {
			msl::Ref ref{ *(this->begin() + begin + j) };
			if (ref.alive()) {
				auto const proba{ get(ref) };
				if (proba > 0) {
					if ((remainder >= eligible_nr) || (unif(generator) < factor * proba)) {
						++n;
						indexes[i].push_back(&ref - &this->front());
						if (--remainder == 0)
							break;
					}
					if (--eligible_nr == 0)
						break;
				}
			}
		}
		sims[i] = n;
	} };
	// Parallelised execution of the task.
	begin = 0;
	for (auto [i, thread] : msl::enumerate(threads)) {
		auto const end{ std::min(begin + step, this->size()) };
		thread = std::thread{ task2, i, begin, end };
		begin = end;
	}
	msl::join(threads);
	auto const sims_nr{ std::accumulate(sims.cbegin(), sims.cend(), size_t{}) };
	if (debug) {
		std::cout << "#\tsum_p\telig\tevent\tsim\n";
		for (auto [i, sum_p] : msl::enumerate(proba_sums))
			std::cout << i << '\t' << sum_p << '\t' << eligibles[i] << '\t' << events_nr[i] << '\t'
			          << sims[i] << '\n';
		std::cout << "Tot" << '\t' << sum_proba << '\t' << eligible_nr << '\t' << n << '\t' << sims_nr
		          << '\n';
	}
	// Last task : execute the user function.
	auto const task3{ [&](size_t i) {
		for (auto const& j : indexes[i])
			function((*this)[j]);
	} };
	for (auto [i, thread] : msl::enumerate(threads))
		thread = std::thread{ task3, i };
	msl::join(threads);
	return sims_nr;
}
size_t
msl::Population::align(size_t n, msl::Table const& get, msl::Function const& function, bool debug)
{
	// A way to check that the variable is a “FLT” one.
	if (get.mis_ == nullptr)
		error_as_interest_variable("align", get.name_());
	return this->align(n, get.flt_, function, debug);
}

// The population is altered,  becoming unusable...
void
msl::Population::save(std::str const& dir_name, std::str const& dir_name_2, bool exit)
{
	msl::TimeWatch tw;
	// Sanity check...
	assert(sizeof(size_t) <= sizeof(msl::Individual*));
	// For each LNK variable, is the sum of the members of each subset equals to the total number
	// of alive individus ?
	size_t const alive{ this->count_if(msl::var::one) };
#define LNK(prm)                                                                                   \
	{                                                                                                \
		size_t total{};                                                                                \
		for (auto const& ref : *this)                                                                  \
			if (ref.prm##_head())                                                                        \
				total += ref.prm##_size();                                                                 \
		assert((total == alive) && #prm);                                                              \
	}
#define FLT(prm)         /* nothing */
#define FLG(prm)         /* nothing */
#define CTG(prm, format) /* nothing */
#define PTR(prm)         /* nothing */
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
	// Map between the address of each individu and his number.
	std::map<msl::Individual*, size_t> ptr2nr;
	// Reserve 0 to denote the end of the linked list.
	ptr2nr[0] = 0;
	for (msl::Ref ref : *this) {
		// No duplicate.
		assert(ptr2nr.find(&ref) == ptr2nr.end());
		// 1 based, not 0 based.
		ptr2nr[&ref] = ptr2nr.size();
	}
	// Replace the address of each individu by its number.
	for (msl::Ref ref : *this) {
#define LNK(prm)                                                                                   \
	assert(ptr2nr.find(ref.prm##_next) != ptr2nr.end());                                             \
	if (ref.prm##_next)                                                                              \
		ref.prm##_next = reinterpret_cast<msl::Individual*>(ptr2nr[ref.prm##_next]);
#define FLT(prm)         /* nothing */
#define FLG(prm)         /* nothing */
#define CTG(prm, format) /* nothing */
#define PTR(prm)                                                                                   \
	assert(ptr2nr.find(ref.prm) != ptr2nr.end());                                                    \
	if (ref.prm)                                                                                     \
		ref.prm = reinterpret_cast<msl::Individual*>(ptr2nr[ref.prm]);
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
	}
	// The task to save a chunk of data.
	auto const task{ [&](std::str const& dir_name, size_t begin, size_t end) {
		if (begin == end)
			return;
		// Compression level 9.
		auto out{ gzopen(dir_name.c_str(), "wb9") };
		if (out == NULL)
			throw msl::Exception{ "unable to open for output the “" + dir_name + "” file" };
		// It is preferable to write individual by individual to not overflow the buffer used to
		// compress the data. When we write in one go, the number of bytes actually written to
		// the file is less than the size of the chunk.
		for (size_t i{ begin }; i < end; ++i) {
			auto const n{ gzwrite(
				out, reinterpret_cast<char const*>(&(*this)[i]), sizeof(msl::Individual)) };
			if (n != sizeof(msl::Individual))
				throw msl::Exception{ "unable to append the zipped stream into the “" + dir_name +
					                    "” file" };
		}
		if (gzclose(out) != Z_OK)
			throw msl::Exception{ "unable to close the “" + dir_name + "” file" };
	} };
	// Create the directory if not exists.
	if (!std::filesystem::exists(dir_name))
		if (!std::filesystem::create_directory(dir_name))
			throw msl::Exception{ "unable to create “" + dir_name + "” directory" };
	if (!dir_name_2.empty())
		if (!std::filesystem::exists(dir_name_2))
			if (!std::filesystem::create_directory(dir_name_2))
				throw msl::Exception{ "unable to create “" + dir_name_2 + "” directory" };

	{
		// The file “info.txt” stores a lot of information ; the “load_population” function will use
		// this information to be sure that the data of the same model is reloaded.
		std::ofstream out{ dir_name + "/info.txt" };
		out << "sizeof(msl::Individual)\t" << sizeof(msl::Individual) << '\n';
		// Build the set of all the variables in the model.
		std::set<std::str> names{
			"alive",
			"birthday",
			"id_number",
#define LNK(prm) #prm "_head", #prm "_next", #prm "_weight",
#define FLT(prm) #prm,
#define FLG(prm) #prm,
#define CTG(prm, format) #prm,
#define PTR(prm) #prm,
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
		};
		out << "names.size()\t" << names.size() << '\n';
		// Put also all the names in the info file.
		for (auto const& name : names)
			out << name << '\n';
		out << "population.size()\t" << this->size() << '\n';
		out << "msl::THREADS_NR\t" << msl::THREADS_NR << '\n';
		auto threads{ msl::reserve<std::thread>(msl::THREADS_NR) };
		auto const step{ (this->size() - 1) / msl::THREADS_NR + 1 };
		size_t begin{};
		for (size_t i{}; i < msl::THREADS_NR; ++i) {
			// The first chunk is named “0.gz”, then “1.gs”, ... If the second file name is filled in,
			// the chunks are distributed over the two directories.
			auto const name{ (dir_name_2.empty() || i % 2)
				                 ? dir_name + '/' + std::to_string(i) + ".gz"
				                 : dir_name_2 + '/' + std::to_string(i) + ".gz" };
			auto const end{ std::min(begin + step, this->size()) };
			out << "begin\t" << begin << '\n';
			out << "end\t" << end << '\n';
			threads.emplace_back(task, name, begin, end);
			begin = end;
		}
		msl::join(threads);
	} // the “out” file is closed.
	std::cout << "msl library message: " << this->size() << " individuals saved (" << alive
	          << " alives) ";
	if (dir_name_2.empty())
		std::cout << "on the directory “" << dir_name << "”.\n";
	else
		std::cout << "on the directories “" << dir_name << "” and “" << dir_name_2 << "”.\n";
	std::cout << "msl library message: elapsed time of the “Population::save” member function: "
	          << tw.elapsed() << " seconds.\n";
	// Call the “exit” function as the data has become unusable.
	if (exit) {
		std::cout << "msl library message: normal program termination after a call to the "
		             "“Population::save” member function."
		          << std::endl;
		std::exit(0);
	}
}
msl::Population
load_population(std::str const& dir_name, std::str const& dir_name_2)
{
	msl::TimeWatch tw;
	size_t n, threads_nr;
	std::vector<std::pair<size_t, size_t>> begin_end;
	{
		std::ifstream in{ dir_name + "/info.txt" };
		assert(in.is_open());
		std::str str;
		in >> str >> n;
		assert("sizeof(msl::Individual)" == str);
		assert(sizeof(msl::Individual) == n);
		// Build the set of all the variables in the model.
		std::set<std::str> names{
			"alive",
			"birthday",
			"id_number",
#define LNK(prm) { #prm "_head" }, { #prm "_next" }, { #prm "_weight" },
#define FLT(prm) { #prm },
#define FLG(prm) { #prm },
#define CTG(prm, format) { #prm },
#define PTR(prm) { #prm },
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
		};
		in >> str >> n;
		assert("names.size()" == str);
		assert(names.size() == n);
		// Read the name of the variables from the info file.
		std::set<std::str> read_names;
		for (size_t i{}; i < n; ++i) {
			in >> str;
			read_names.insert(str);
		}
		// Check if it is the same set of variables.
		assert(names == read_names);
		in >> str >> n;
		assert("population.size()" == str);
		in >> str >> threads_nr;
		assert("msl::THREADS_NR" == str);
		if (threads_nr != msl::THREADS_NR)
			std::cerr << "msl library warning: different count of cores...\n";
		// Read the begin and the end indices of the each chunk.
		for (size_t i{}; i < threads_nr; ++i) {
			size_t begin, end;
			in >> str >> begin;
			assert("begin" == str);
			in >> str >> end;
			assert("end" == str);
			begin_end.push_back({ begin, end });
		}
		// Check that the end of the previous chunk is equal to the begin of the current chunk.
		for (size_t i{ 1 }; i < threads_nr; ++i)
			assert(begin_end[i - 1].second == begin_end[i].first);
	}
	// The population to fill, initialised to not a number.
	msl::Population rvo(n);
	// The task for reading and decompressing a data chunk.
	auto const task{ [&](std::str const& dir_dir_name, size_t begin, size_t end) {
		if (begin == end)
			return;
		auto in{ gzopen(dir_dir_name.c_str(), "rb") };
		if (in == NULL)
			throw msl::Exception{ "unable to open for input the “" + dir_dir_name + "” file" };
		// Before, the chunk was read in one go :
		//   gzread(in, reinterpret_cast<char*>(&rvo[begin]), (end-begin) * sizeof(msl::Individual));
		// I suspect that the transfert of the data and the decompress process was not parallelised.
		// The speedup is 15 % with the read by individual.
		for (size_t i{ begin }; i < end; ++i)
			if (gzread(in, reinterpret_cast<char*>(&rvo[i]), sizeof(msl::Individual)) !=
			    sizeof(msl::Individual))
				throw msl::Exception{ "unable to read the zipped stream from the “" + dir_dir_name +
					                    "” file" };
		if (gzclose(in) != Z_OK)
			throw msl::Exception{ "unable to close the “" + dir_dir_name + "” file" };
	} };
	// Read and decompress each chunk in parallel.
	auto threads{ msl::reserve<std::thread>(threads_nr) };
	for (size_t i{}; i < threads_nr; ++i) {
		auto const dir_dir_name{ (dir_name_2.empty() || i % 2)
			                         ? dir_name + '/' + std::to_string(i) + ".gz"
			                         : dir_name_2 + '/' + std::to_string(i) + ".gz" };
		threads.emplace_back(task, dir_dir_name, begin_end[i].first, begin_end[i].second);
	}
	msl::join(threads);
	// Replace the individu number by its pointer.
	for (auto& ref : rvo) {
#define LNK(prm)                                                                                   \
	if (ref.prm##_next) {                                                                            \
		auto const n{ reinterpret_cast<size_t>(ref.prm##_next) };                                      \
		if (n > rvo.size()) {                                                                          \
			MSL_SHOW(&ref - &rvo[0])                                                                     \
			MSL_SHOW(n)                                                                                  \
			MSL_SHOW(rvo.size())                                                                         \
			MSL_SHOW(#prm)                                                                               \
			assert(n <= rvo.size());                                                                     \
		}                                                                                              \
		if (n > 0)                                                                                     \
			ref.prm##_next = &rvo[n - 1];                                                                \
	}
#define FLT(prm)         /* nothing */
#define FLG(prm)         /* nothing */
#define CTG(prm, format) /* nothing */
#define PTR(prm)                                                                                   \
	if (ref.prm) {                                                                                   \
		auto const n{ reinterpret_cast<size_t>(ref.prm) };                                             \
		assert(n <= rvo.size());                                                                       \
		if (n > 0)                                                                                     \
			ref.prm = &rvo[n - 1];                                                                       \
	}
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
	}
	auto const alive{ rvo.count_if(msl::var::one) };
	std::cout << "msl library message: " << rvo.size() << " individuals loaded (" << alive
	          << " alives) ";
	if (dir_name_2.empty())
		std::cout << "from the directory “" << dir_name << "”.\n";
	else
		std::cout << "from the directories “" << dir_name << "” and “" << dir_name_2 << "”.\n";
	std::cout << "msl library message: elapsed time of the “load_population” function: "
	          << tw.elapsed() << " seconds.\n";
	return rvo;
}
// The population is altered,  becoming unusable...
void
msl::Population::transport_save(std::str const& file_name, bool exit)
{
	msl::TimeWatch tw;
	// Sanity check...
	assert(sizeof(size_t) <= sizeof(msl::Individual*));
	size_t const alive{ this->count_if(msl::var::one) };
	// For each LNK variable, is the sum of the members of each subset equals to the total number
	// of individus ?
#define LNK(prm)                                                                                   \
	{                                                                                                \
		size_t total{};                                                                                \
		for (msl::RefConst ref : *this)                                                                \
			if (ref.prm##_head())                                                                        \
				total += ref.prm##_size();                                                                 \
		assert(total == alive);                                                                        \
	}
#define FLT(prm)         /* nothing */
#define FLG(prm)         /* nothing */
#define CTG(prm, format) /* nothing */
#define PTR(prm)         /* nothing */
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
	std::map<msl::Individual*, size_t> ptr2nr;
	// Reserve 0 to denote the end of the linked list.
	ptr2nr[0] = 0;
	for (msl::Ref ref : *this) {
		// No duplicate.
		assert(ptr2nr.find(&ref) == ptr2nr.end());
		// 1 based, not 0 based.
		ptr2nr[&ref] = ptr2nr.size();
	}
	// Replace the address of each individu by its number.
	for (msl::Ref ref : *this) {
#define LNK(prm)                                                                                   \
	assert(ptr2nr.find(ref.prm##_next) != ptr2nr.end());                                             \
	if (ref.prm##_next)                                                                              \
		ref.prm##_next = reinterpret_cast<msl::Individual*>(ptr2nr[ref.prm##_next]);
#define FLT(prm)         /* nothing */
#define FLG(prm)         /* nothing */
#define CTG(prm, format) /* nothing */
#define PTR(prm)                                                                                   \
	assert(ptr2nr.find(ref.prm) != ptr2nr.end());                                                    \
	if (ref.prm)                                                                                     \
		ref.prm = reinterpret_cast<msl::Individual*>(ptr2nr[ref.prm]);
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
	}
	fd_write_xlsx::sheet_t sheet;
	// The first row is for the variable names.
	{
		fd_write_xlsx::row_t row;
		row.push_back({ "alive" });
		row.push_back({ "birthday" });
		row.push_back({ "id_number" });
#define LNK(prm)                                                                                   \
	row.push_back({ #prm "_head" });                                                                 \
	row.push_back({ #prm "_next" });                                                                 \
	row.push_back({ #prm "_weight" });
#define FLT(prm) row.push_back({ #prm });
#define FLG(prm) row.push_back({ #prm });
#define CTG(prm, format) row.push_back({ #prm });
#define PTR(prm) row.push_back({ #prm });
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
		sheet.push_back(row);
	}
	// The other rows are for data; integer values become “int64_t” values using explicit
	// conversion.
	for (msl::RefConst ref : *this) {
		std::vector<fd_write_xlsx::cell_t> row;
		row.push_back({ int64_t(ref.alive()) });
		row.push_back({ int64_t(ref.birthday) });
		row.push_back({ int64_t(ref.id_number) });
#define LNK(prm)                                                                                   \
	row.push_back({ int64_t(ref.prm##_head()) });                                                    \
	row.push_back({ int64_t(ref.prm##_next) });                                                      \
	row.push_back({ ref.prm##_weight });
#define FLT(prm) row.push_back({ ref.prm });
#define FLG(prm) row.push_back({ int64_t(ref.prm()) });
#define CTG(prm, format) row.push_back({ int64_t(ref.prm) });
#define PTR(prm) row.push_back({ int64_t(ref.prm) });
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
		sheet.push_back(row);
	}
	fd_write_xlsx::write(file_name, sheet);
	std::cout
	  << "msl library message: elapsed time of the “Population::transport_save” member function: "
	  << tw.elapsed() << " seconds.\n";
	// Call the “exit” function as the data has become unusable.
	if (exit) {
		std::cout << "msl library message: normal program termination after a call to the "
		             "“Population::transport_save” member function."
		          << std::endl;
		std::exit(0);
	}
}

msl::Population
transport_load_population(std::str const& file_name)
{
	msl::TimeWatch tw;
	auto const no_col{ std::numeric_limits<size_t>::max() };
	auto const sheet{ fd_read_xlsx::read(file_name) };
	// Build a map whose key is the name variable and value is the column number in the sheet; the
	// value is initialised to “no_col”.
	std::map<std::str, size_t> name2col{
		{ "alive", no_col },
		{ "birthday", no_col },
		{ "id_number", no_col },
#define LNK(prm) { #prm "_head", no_col }, { #prm "_next", no_col }, { #prm "_weight", no_col },
#define FLT(prm) { #prm, no_col },
#define FLG(prm) { #prm, no_col },
#define CTG(prm, format) { #prm, no_col },
#define PTR(prm) { #prm, no_col },
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
	};
	// The first row contains the name of the variables in the sheet.
	for (size_t j{}; j < sheet[0].size(); ++j) {
		if (fd_read_xlsx::holds_string(sheet[0][j])) {
			auto const& name{ fd_read_xlsx::get_string(sheet[0][j]) };
			auto const iterator{ name2col.find(name) };
			// The name is a variable name in the model; update the map value with the column number.
			if (iterator != cend(name2col))
				iterator->second = j;
			else
				std::cerr << "msl library warning: “" << iterator->first
				          << "” variable not in the model...\n";
		}
	}
	// Display a warning if some variables in the model are not in the sheet.
	for (auto const& [key, value] : name2col)
		if (value == no_col)
			std::cerr << "msl library warning: “" << key << "” variable not in the transport file...\n";
	auto const n{ sheet.size() - 1 };
	// The population to fill.
	msl::Population rvo(n);
	{
		auto const iter{ name2col.find("alive") };
		if (iter->second != no_col)
			for (size_t i{}; i < n; ++i)
				rvo[i].set_alive(fd_read_xlsx::get_int(sheet[i + 1][iter->second]));
	}
	{
		auto const iter{ name2col.find("birthday") };
		if (iter->second != no_col)
			for (size_t i{}; i < n; ++i)
				rvo[i].set_birthday(fd_read_xlsx::get_int(sheet[i + 1][iter->second]));
	}
	{
		auto const iter{ name2col.find("id_number") };
		if (iter->second != no_col)
			for (size_t i{}; i < n; ++i)
				rvo[i].set_id_number(fd_read_xlsx::get_int(sheet[i + 1][iter->second]));
	}
	// Replace the individu number by its pointer for the linked lists and for the “PTR” variables.
#define LNK(prm)                                                                                   \
	{                                                                                                \
		auto const iter{ name2col.find(#prm "_head") };                                                \
		if (iter->second != no_col)                                                                    \
			for (size_t i{}; i < n; ++i)                                                                 \
				rvo[i].set_##prm##_head(fd_read_xlsx::get_int(sheet[i + 1][iter->second]));                \
	}                                                                                                \
	{                                                                                                \
		auto const iter{ name2col.find(#prm "_next") };                                                \
		if (iter->second != no_col)                                                                    \
			for (size_t i{}; i < n; ++i) {                                                               \
				auto const m{ fd_read_xlsx::get_int(sheet[i + 1][iter->second]) };                         \
				assert((m >= 0) && (size_t(m) <= rvo.size()));                                             \
				if (m > 0)                                                                                 \
					rvo[i].prm##_next = &rvo[m - 1];                                                         \
			}                                                                                            \
	}                                                                                                \
	{                                                                                                \
		auto const iter{ name2col.find(#prm "_weight") };                                              \
		if (iter->second != no_col)                                                                    \
			for (size_t i{}; i < n; ++i)                                                                 \
				rvo[i].set_##prm##_weight(fd_read_xlsx::get_double(sheet[i + 1][iter->second]));           \
	}
#define FLT(prm)                                                                                   \
	{                                                                                                \
		auto const iter{ name2col.find(#prm) };                                                        \
		if (iter->second != no_col)                                                                    \
			for (size_t i{}; i < n; ++i)                                                                 \
				rvo[i].set_##prm(fd_read_xlsx::get_double(sheet[i + 1][iter->second]));                    \
	}
#define FLG(prm)                                                                                   \
	{                                                                                                \
		auto const iter{ name2col.find(#prm) };                                                        \
		if (iter->second != no_col)                                                                    \
			for (size_t i{}; i < n; ++i)                                                                 \
				rvo[i].set_##prm(fd_read_xlsx::get_int(sheet[i + 1][iter->second]));                       \
	}
#define CTG(prm, format)                                                                           \
	{                                                                                                \
		auto const iter{ name2col.find(#prm) };                                                        \
		if (iter->second != no_col)                                                                    \
			for (size_t i{}; i < n; ++i)                                                                 \
				rvo[i].set_##prm(fd_read_xlsx::get_int(sheet[i + 1][iter->second]));                       \
	}
#define PTR(prm)                                                                                   \
	{                                                                                                \
		auto const iter{ name2col.find(#prm) };                                                        \
		if (iter->second != no_col)                                                                    \
			for (size_t i{}; i < n; ++i) {                                                               \
				auto const m{ fd_read_xlsx::get_int(sheet[i + 1][iter->second]) };                         \
				assert((m >= 0) && (size_t(m) <= rvo.size()));                                             \
				rvo[i].prm = (m > 0) ? &rvo[m - 1] : nullptr;                                              \
			}                                                                                            \
	}
#include "msl-variables.cpp"
#undef PTR
#undef CTG
#undef FLG
#undef FLT
#undef LNK
	std::cout << "msl library message: elapsed time of the “transport_load_population” function: "
	          << tw.elapsed() << " seconds.\n";
	return rvo;
}
// Load from a xlsx workbook the parameters of the model selecting a column for a year.
void
load_parameters(std::str const& file_name, size_t year)
{
	// Load the active sheet of the workbook.
	auto const sheet{ fd_read_xlsx::read(file_name) };
	// Search for the column with the good “year” value on first row.
	auto const col{ [&]() {
		for (size_t j{}; j < sheet[0].size(); ++j)
			if (fd_read_xlsx::compare(sheet[0][j], int64_t(year)))
				return j;
		throw Exception{ "unable to find “" + std::to_string(year) + "” column in active sheet of “" +
			               file_name + "” workbook" };
	}() };
	// Set the parameter value using the “prm2ptr_” map. The name of the parameter is in the first
	// column.
	for (size_t i{ 1 }; i < sheet.size(); ++i) {
		if ((col < sheet[i].size()) && fd_read_xlsx::holds_string(sheet[i][0]) &&
		    fd_read_xlsx::holds_num(sheet[i][col])) {
			auto const& prm_name{ fd_read_xlsx::get_string(sheet[i][0]) };
			auto const iterator{ prm::prm2ptr_.find(prm_name) };
			if (iterator == cend(prm::prm2ptr_))
				std::cerr << "msl library warning: parameter “" << prm_name
				          << "” not found in the model...\n";
			else
				*(iterator->second) = fd_read_xlsx::get_num(sheet[i][col]);
		}
	}
	// Display a warning for the parameters not setted.
	for (auto const& [key, value] : prm::prm2ptr_)
		if (std::isnan(*value))
			std::cerr << "msl library warning: parameter “" << key << "” not found in the workbook...\n";
}
// Load the formats from a xlsx workbook, each format being on a worksheet.
void
load_formats(std::str const& file_name)
{
	// For each “CTG” variable in the model.
	for (auto& [first, second] : global_formats) {
		auto const& format_name{ second.first };
		auto& map{ second.second };
		try {
			// Try to read the sheet whose name is the format name of the “CTG” variable.
			auto const sheet{ fd_read_xlsx::read(file_name, format_name.c_str()) };
			for (size_t i{}; i < sheet.size(); ++i) {
				// The first column is the value of the “CTG” variable and the second column is the label.
				if ((1 < sheet[i].size()) && fd_read_xlsx::holds_num(sheet[i][0]) &&
				    fd_read_xlsx::holds_string(sheet[i][1])) {
					auto const& key{ fd_read_xlsx::get_num(sheet[i][0]) };
					// The key must be unique.
					assert(map.find(uint8_t(key)) == cend(map));
					// The value is the label, a string value.
					auto const& value{ fd_read_xlsx::get_string(sheet[i][1]) };
					map[uint8_t(key)] = value;
				}
			}
		} catch (...) {
			std::cerr << "msl library warning: unable to get “" + format_name + "” format in the “" +
			               file_name + "” workbook."
			          << std::endl;
		}
	}
}
// Save in a xlxs worbook the map returned by the “sum_by” function. The key of this map is a
// value of the by variable. The value of this map is a pair : first, the label of the by variable
// value and, second, the sum of the variable of interest for the by variable value.
void
Tabular::save(std::str const& file_name) const
{
	fd_write_xlsx::sheet_t sheet;
	// The first row is used to name the columns.
	sheet.push_back({
	  { "Ctg" },
	  { "Label" },
	  { "Sum" },
	});

	for (auto const& [key, value] : *this)
		sheet.push_back({ { int64_t(key) }, { value.first }, { value.second } });

	// The last row is used to put the sum of the “sum” column.
	sheet.push_back({
	  {},
	  { "Total" },
	  { std::accumulate(this->cbegin(),
	                    this->cend(),
	                    double{},
	                    [](double cumul, auto pair) { return cumul + pair.second.second; }) },
	});
	fd_write_xlsx::write(file_name, sheet);
}
// Save in a xlxs worbook the maps returned by two calls of the “sum_by” function. The key of
// those maps is a value of the by variable. The value of those maps is a pair : first, the label
// of the by variable value and, second, the sum of the variable of interest for the by variable
// value.
void
Tabular::save(Tabular const& tab, std::str const& file_name) const
{
	fd_write_xlsx::sheet_t sheet;
	sheet.push_back({
	  { "By" },
	  { "Label" },
	  { "Sum 1" },
	  { "Sum 2" },
	  { "Delta" },
	  { "Delta %" },
	});
	std::set<size_t> keys;
	std::transform(this->cbegin(), this->cend(), inserter(keys, keys.begin()), [](auto pair) {
		return pair.first;
	});
	std::transform(
	  tab.cbegin(), tab.cend(), inserter(keys, keys.begin()), [](auto pair) { return pair.first; });
	double sum1{}, sum2{};
	for (auto const& key : keys) {
		auto const iterator1{ this->find(key) };
		auto const iterator2{ tab.find(key) };
		if (iterator1 == this->cend()) {
			sheet.push_back(
			  { { int64_t(key) }, { iterator2->second.first }, { "" }, { iterator2->second.second } });
			sum2 += iterator2->second.second;
		} else if (iterator2 == tab.cend()) {
			sheet.push_back(
			  { { int64_t(key) }, { iterator1->second.first }, { iterator1->second.second }, { "" } });
			sum1 += iterator1->second.second;
		} else {
			// It is the opportunity to check that it is the same by variable for the two maps.
			assert(iterator1->second.first == iterator2->second.first);
			sheet.push_back({ { int64_t(key) },
			                  { iterator1->second.first },
			                  { iterator1->second.second },
			                  { iterator2->second.second },
			                  { iterator2->second.second - iterator1->second.second },
			                  { (iterator1->second.second != 0)
			                      ? 100 * (iterator2->second.second - iterator1->second.second) /
			                          iterator1->second.second
			                      : std::numeric_limits<double>::quiet_NaN() } });
			sum1 += iterator1->second.second;
			sum2 += iterator2->second.second;
		}
	}
	sheet.push_back({
	  {},
	  { "Total" },
	  { sum1 },
	  { sum2 },
	  { sum2 - sum1 },
	  { (sum1 != 0) ? 100 * (sum2 - sum1) / sum1 : std::numeric_limits<double>::quiet_NaN() },
	});
	fd_write_xlsx::write(file_name, sheet);
}

} // namespace msl
